# Development Guide

This article will help you build, test and consume local builds of the MSTest Framework and Adapter.

## Prerequisites

Please install [Visual Studio 2017](https://www.microsoft.com/net/core#windowsvs2017) with the following workloads:

- `.NET desktop development`
- `Universal Windows Platform development`
- `.Net Core cross-platform development`

## Recommended workflow

We recommend the following overall workflow when developing for this repository:

- Fork this repository.
- Always work in your fork.
- Always keep your fork up to date.

### Starting with your fork

Before updating your fork, run this command:

```shell
git remote add upstream https://github.com/Microsoft/testfx.git
```

This will make management of multiple forks and your own work easier over time.

### Updating your fork

We recommend the following commands to update your fork:

```shell
git checkout main
git clean -dfx
git fetch upstream
git rebase upstream/main
git push
```

Or more succinctly:

```shell
git checkout main && git clean -xdf && git fetch upstream && git rebase upstream/main && git push
```

This will update your fork with the latest from `microsoft/testfx` on your machine and push those updates to your remote fork.

## Building

Lets assume `/src/testfx` as the location of the cloned repository for the rest of this article.

### Building with Visual Studio (windows only)

You can open `/src/testfx/TestFx.sln` in VS and trigger a build of the entire code base using `Build Solution (Ctrl+Shift+B)` from the `Solution Explorer` or the `Build` menu.

The artifacts get dropped at `/src/testfx/artifacts/<Configuration>/` (e.g., `/src/testfx/artifacts/Debug/` for `Debug` mode - default).

### Building with command line (CLI)

To build the repository, run the following command:

```shell
cd /src/testfx
./build.cmd
```

This would use the msbuild that Visual Studio installation brings in to build the following components:

- The Framework and its inbox extensions.
- The Adapter and its platform specific components.
- Unit, Component and E2E tests related to the above components.
- Generates a NuGet package for local consumption of the Framework and Adapter.

All these components get dropped at `/src/testfx/artifacts/<Configuration>/` and the NuGet packages would be at `/src/testfx/artifacts/<Configuration>/MSTestPackages`.

You can configure the build using build flags.

To build a particular configuration, use the `-c` option. E.g. to trigger a
release build use

```shell
./build.cmd -c Release
```

To change the version suffix of the nuget packages generated, a `-vs` parameter can be passed through as follows (By default this is **dev**)

```shell
./build.cmd -vs dev-01
```

For more options, you can use

```shell
./build.cmd -h
```

## Testing

The following are the set of tests that `testfx` contains:

- Unit tests
  - Very fast tests primarily validating individual units.
  - Named as `<ProjectUnderTest>.UnitTests` where ProjectUnderTest is the project under test.
- Component tests
  - Slightly slower tests with File system interactions.
  - Named as `<ProjectUnderTest>.ComponentTests` where ProjectUnderTest is the project under test.
- Smoke tests
  - End to end tests covering P0 workflows which most users would use. if these are broken, PR will not be merged.
  - Run using the Framework and adapter bits generated by the build.
  - Named as `MSTestAdapter.Smoke.E2ETests`

As a principle, the test bed would consist mostly of unit tests (~70-80%), followed by a few component tests addressing real world interactions (~15-20%) and a few end to end tests (~5-10%).

### Testing with Visual Studio (windows only)

All the tests in the testfx repo can be run via the `Visual Studio Test Explorer`. Building `/src/testfx/TestFx.sln` as described in the build section above should populate all the tests in the Test Explorer.

A specific type of tests can be run by providing a search filter in the Test Explorer window.
For instance to run unit tests use `project:"Unit"`. For running smoke tests, use the `project:"Smoke"` filter.

![image](https://user-images.githubusercontent.com/11340282/194912284-88718bb6-ce20-492a-adab-15d032f6bb45.png)

### Running tests with command line (CLI)

To execute tests via command line, run the following command:

```shell
cd /src/testfx
test.cmd
```

By default, only unit tests are run. To run smoke tests, one can provide the `-p` option to `test.cmd` that sets the test assembly pattern:

```shell
test.cmd -p smoke
```

The `-p` option can also be used to run tests from a specific assembly. For instance to run *TestFramework* tests the following command can be used:

```shell
test.cmd -p TestFramework
```

Tests can also be run for a specific build configuration (`Debug` being the default)

```shell
test.cmd -c release
```

## Deploying

This section will discuss the steps to consume the locally built Framework and adapter bits that the sections above detail.

On running `build.cmd`, language neutral NuGet packages for the Framework and Adapter with a version of `99.99.99-dev` are generated at `src\testfx\artifacts\MSTestPackages`.

### Deploying in Visual Studio

A test project can be updated to consume these NuGet packages by:

1. Adding `src\testfx\artifacts\MSTestPackages` to the list of Package sources in VS via `Tools-> Options -> Nuget Package Manager -> Package Sources`.
2. Updating the versions of these nuget packages to point to `99.99.99-dev` for the test project/solution via the `Manage Nuget Package` workflow at a project/solution level.

**Note:** Owing to a caching issue in VS Test Explorer, please restart VS to ensure that these packages are actually consumed.

### Deploying through CLI/non-VS IDEs

These packages can be consumed from a non-VS IDE/editor by:

1. Updating their versions to point to `99.99.99-dev` for the test project/solution using:

```shell
nuget.exe update -id MSTest.TestFramework -id MSTest.TestAdapter -version 99.99.99-dev -source "<Root>\src\testfx\artifacts\Debug\MSTestPackages" -prerelease packages.config
```

Documentation for nuget.exe command line reference is [here](https://docs.microsoft.com/nuget/tools/nuget-exe-cli-reference)

## Diagnostics

The first level of diagnosis for adapter failures can start with enabling verbose logging in the Visual studio Test Platform itself. Here is how to turn that on:

- **TP V1** : [This](https://blogs.msdn.microsoft.com/aseemb/2012/03/01/how-to-enable-ute-logs/) blog helps detail setting up diagnostic logging.
- **TP V2**(The open-source cross-plat Test Platform): [This](https://github.com/Microsoft/vstest-docs/blob/main/docs/diagnose.md#collect-traces-using-command-line) section helps detail the process to enable diagnostic logging.

One can always add a `Debugger.Launch` at the main entry points:

- [MSTestDiscoverer.DiscoverTests](https://github.com/Microsoft/testfx/blob/main/src/Adapter/MSTest.CoreAdapter/MSTestDiscoverer.cs) for discovery.
- [MSTestExecutor.RunTests](https://github.com/Microsoft/testfx/blob/main/src/Adapter/MSTest.CoreAdapter/MSTestExecutor.cs) - both the overloads that take sources and tests for execution.
 Select the appropriate debugger and step through the code.
